"""Define /user endpoints."""
from typing import Awaitable, Callable

from .const import API_V2_BASE


class User:  # pylint: disable=too-few-public-methods
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable], user_id: str) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request
        self._user_id: str = user_id

    async def get_info(
        self, include_alarm_settings: bool = False, include_location_info: bool = False
    ) -> dict:
        """Return user account data.

        :param include_alarm_settings: Include expanded alarm information
        :type include_alarm_settings: ``bool``
        :param include_location_info: Include expanded location info
        :type include_location_info: ``bool``
        :rtype: ``dict``
        """
        additional_info = []
        if include_alarm_settings:
            additional_info.append("alarmSettings")
        if include_location_info:
            additional_info.append("locations")

        params = {}
        if additional_info:
            params["expand"] = ",".join(additional_info)

        return await self._request(
            "get", f"{API_V2_BASE}/users/{self._user_id}", params=params
        )



"""Define /device endpoints."""
from typing import Awaitable, Callable

from .const import API_V2_BASE


class Device:  # pylint: disable=too-few-public-methods
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request

    async def get_info(self, device_id: str) -> dict:
        """Return device specific data.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request("get", f"{API_V2_BASE}/devices/{device_id}")

    async def run_health_test(self, device_id: str) -> None:
        """Run a health test for a specific device.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request(
            "post", f"{API_V2_BASE}/devices/{device_id}/healthTest/run"
        )

    async def open_valve(self, device_id: str) -> None:
        """Open the valve for a specific device.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request(
            "post",
            f"{API_V2_BASE}/devices/{device_id}",
            json={"valve": {"target": "open"}},
        )

    async def close_valve(self, device_id: str) -> None:
        """Close the valve for a specific device.

        :param device_id: Unique identifier for the device
        :type device_id: ``str``
        :rtype: ``dict``
        """
        return await self._request(
            "post",
            f"{API_V2_BASE}/devices/{device_id}",
            json={"valve": {"target": "closed"}},
        )



"""Define /water endpoints."""
from datetime import datetime
from typing import Awaitable, Callable

from .const import API_V2_BASE
from .util import raise_on_invalid_argument

INTERVAL_DAILY = "1d"
INTERVAL_HOURLY = "1h"
INTERVAL_MONTHLY = "1m"
INTERVALS = set([INTERVAL_DAILY, INTERVAL_HOURLY, INTERVAL_MONTHLY])


class Water:  # pylint: disable=too-few-public-methods
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request

    async def get_consumption_info(
        self,
        location_id: str,
        start: datetime,
        end: datetime,
        interval: str = INTERVAL_HOURLY,
    ) -> dict:
        """Return user account data.

        :param location_id: A Flo location UUID
        :type location_id: ``str``
        :param start: The start datetime of the range to examine
        :type start: ``datetime.datetime``
        :param end: The end datetime of the range to examine
        :type end: ``datetime.datetime``
        :rtype: ``dict``
        """
        raise_on_invalid_argument(interval, INTERVALS)

        return await self._request(
            "get",
            f"{API_V2_BASE}/water/consumption",
            params={
                "endDate": end.isoformat(),
                "interval": interval,
                "locationId": location_id,
                "startDate": start.isoformat(),
            },
        )

    async def get_metrics(
        self,
        device_mac_address: str,
        start: datetime,
        end: datetime,
        interval: str = INTERVAL_HOURLY,
    ) -> dict:
        """Return user account data.

        :param start: The start datetime of the range to examine
        :type start: ``datetime.datetime``
        :param end: The end datetime of the range to examine
        :type end: ``datetime.datetime``
        :rtype: ``dict``
        """
        raise_on_invalid_argument(interval, INTERVALS)

        return await self._request(
            "get",
            f"{API_V2_BASE}/water/metrics",
            params={
                "endDate": end.isoformat(),
                "interval": interval,
                "macAddress": device_mac_address.replace(":", ""),
                "startDate": start.isoformat(),
            },
        )



"""Define a base client for interacting with Flo."""
from datetime import datetime
import logging
from typing import Optional
from urllib.parse import urlparse

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from .alarm import Alarm
from .device import Device
from .errors import RequestError
from .location import Location
from .user import User
from .water import Water

_LOGGER = logging.getLogger(__name__)

API_V1_BASE: str = "https://api.meetflo.com/api/v1"

DEFAULT_HEADER_ACCEPT: str = "application/json, text/plain, */*"
DEFAULT_HEADER_CONTENT_TYPE: str = "application/json;charset=UTF-8"
DEFAULT_HEADER_ORIGIN: str = "https://user.meetflo.com"
DEFAULT_HEADER_REFERER: str = "https://user.meetflo.com/home"
DEFAULT_HEADER_USER_AGENT: str = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_2) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36"
)
DEFAULT_TIMEOUT: int = 10


class API:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Define the API object."""

    def __init__(
        self, username: str, password: str, *, session: Optional[ClientSession] = None
    ) -> None:
        """Initialize."""
        self._password: str = password
        self._session: ClientSession = session
        self._token: Optional[str] = None
        self._token_expiration: Optional[datetime] = None
        self._user_id: Optional[str] = None
        self._username: str = username

        self.alarm: Alarm = Alarm(self._request)
        self.location: Location = Location(self._request)
        self.water: Water = Water(self._request)
        self.device: Device = Device(self._request)

        # These endpoints will get instantiated post-authentication:
        self.user: Optional[User] = None

    async def _request(self, method: str, url: str, **kwargs) -> dict:
        """Make a request against the API."""
        if self._token_expiration and datetime.now() >= self._token_expiration:
            _LOGGER.info("Requesting new access token to replace expired one")

            # Nullify the token so that the authentication request doesn't use it:
            self._token = None

            # Nullify the expiration so the authentication request doesn't get caught
            # here:
            self._token_expiration = None

            await self.async_authenticate()

        kwargs.setdefault("headers", {})
        kwargs["headers"].update(
            {
                "Accept": DEFAULT_HEADER_ACCEPT,
                "Content-Type": DEFAULT_HEADER_CONTENT_TYPE,
                "Host": urlparse(url).netloc,
                "Origin": DEFAULT_HEADER_ORIGIN,
                "Referrer": DEFAULT_HEADER_REFERER,
                "User-Agent": DEFAULT_HEADER_USER_AGENT,
            }
        )

        if self._token:
            kwargs["headers"]["Authorization"] = self._token

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        try:
            async with session.request(method, url, **kwargs) as resp:
                data: dict = await resp.json(content_type=None)
                resp.raise_for_status()
                return data
        except ClientError as err:
            raise RequestError(f"There was an error while requesting {url}") from err
        finally:
            if not use_running_session:
                await session.close()

    async def async_authenticate(self) -> None:
        """Authenticate the user and set the access token with its expiration."""
        auth_response: dict = await self._request(
            "post",
            f"{API_V1_BASE}/users/auth",
            json={"username": self._username, "password": self._password},
        )

        self._token = auth_response["token"]
        self._token_expiration = datetime.fromtimestamp(
            auth_response["tokenPayload"]["timestamp"]
            + auth_response["tokenExpiration"]
        )

        if not self._user_id:
            self._user_id = auth_response["tokenPayload"]["user"]["user_id"]
            assert self._user_id
            self.user = User(self._request, self._user_id)


async def async_get_api(
    username: str, password: str, *, session: Optional[ClientSession] = None
) -> API:
    """Instantiate an authenticated API object.

    :param session: An ``aiohttp`` ``ClientSession``
    :type session: ``aiohttp.client.ClientSession``
    :param email: A Flo email address
    :type email: ``str``
    :param password: A Flo password
    :type password: ``str``
    :rtype: :meth:`aioflo.api.API`
    """
    api = API(username, password, session=session)
    await api.async_authenticate()
    return api



"""Define general utilities."""
from ..errors import RequestError


def raise_on_invalid_argument(value, options):
    """Raise a RequestError when an invalid consumption interval is provided."""
    if value not in options:
        raise RequestError(
            f"Invalid keyword argument: {value} (valid options: {options})"
        )



"""Define /location endpoints."""
from typing import Awaitable, Callable, Optional

from .const import API_V2_BASE
from .util import raise_on_invalid_argument

SYSTEM_MODE_AWAY = "away"
SYSTEM_MODE_HOME = "home"
SYSTEM_MODE_SLEEP = "sleep"
SYSTEM_MODES = set([SYSTEM_MODE_AWAY, SYSTEM_MODE_HOME, SYSTEM_MODE_SLEEP])
SYSTEM_REVERT_MODES = set([SYSTEM_MODE_AWAY, SYSTEM_MODE_HOME])

# The Flo web app hardcodes sleep durations to certain intervals; testing seems to
# indicate that those are the only valid values, so we stick with them:
SLEEP_MINUTE_OPTIONS = set([120, 1440, 4320])


class Location:
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request

    async def _set_system_mode(
        self, location_id: str, mode: str, additional_payload: Optional[dict] = None
    ) -> None:
        """Set the system mode (with optional parameters)."""
        raise_on_invalid_argument(mode, SYSTEM_MODES)

        payload = {"target": mode}
        if additional_payload:
            payload = {**payload, **additional_payload}
        await self._request(
            "post", f"{API_V2_BASE}/locations/{location_id}/systemMode", json=payload
        )

    async def get_info(
        self, location_id: str, include_device_info: bool = False,
    ) -> dict:
        """Return user account data.

        :param location_id: A Flo location UUID
        :type location_id: ``str``
        :param include_device_info: Include expanded device information
        :type include_device_info: ``bool``
        :rtype: ``dict``
        """
        additional_info = []
        if include_device_info:
            additional_info.append("devices")

        params = {}
        if additional_info:
            params["expand"] = ",".join(additional_info)

        return await self._request(
            "get", f"{API_V2_BASE}/locations/{location_id}", params=params
        )

    async def set_mode_away(self, location_id: str) -> None:
        """Set the system mode to "Away".

        :param location_id: A Flo location UUID
        :type location_id: ``str``
        """
        await self._set_system_mode(location_id, SYSTEM_MODE_AWAY)

    async def set_mode_home(self, location_id: str) -> None:
        """Set the system mode to "Home".

        :param location_id: A Flo location UUID
        :type location_id: ``str``
        """
        await self._set_system_mode(location_id, SYSTEM_MODE_HOME)

    async def set_mode_sleep(
        self,
        location_id: str,
        revert_minutes: int,
        revert_mode: Optional[str] = SYSTEM_MODE_HOME,
    ) -> None:
        """Set the system mode to "Home".

        :param location_id: A Flo location UUID
        :type location_id: ``str``
        :param revert_minutes: The number of minutes to sleep (120, 1440, or 4320)
        :type revert_minutes: ``int``
        :param revert_mode: The mode to set after sleep concludes ("away" or "home")
        :type revert_mode: ``str``
        """
        raise_on_invalid_argument(revert_minutes, SLEEP_MINUTE_OPTIONS)
        raise_on_invalid_argument(revert_mode, SYSTEM_REVERT_MODES)

        await self._set_system_mode(
            location_id,
            SYSTEM_MODE_SLEEP,
            additional_payload={
                "revertMinutes": revert_minutes,
                "revertMode": revert_mode,
            },
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aioflo', 'aioflo.util']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'aioflo',
    'version': '2021.10.0',
    'description': 'A Python3, async-friendly library for Flo by Moen Smart Water Detectors',
    'long_description': '# 💧 aioflo: a Python3, asyncio-friendly library for Flo Smart Water Detectors\n\n[![CI](https://github.com/bachya/aioflo/workflows/CI/badge.svg)](https://github.com/bachya/aioflo/actions)\n[![PyPi](https://img.shields.io/pypi/v/aioflo.svg)](https://pypi.python.org/pypi/aioflo)\n[![Version](https://img.shields.io/pypi/pyversions/aioflo.svg)](https://pypi.python.org/pypi/aioflo)\n[![License](https://img.shields.io/pypi/l/aioflo.svg)](https://github.com/bachya/aioflo/blob/master/LICENSE)\n[![Code Coverage](https://codecov.io/gh/bachya/aioflo/branch/master/graph/badge.svg)](https://codecov.io/gh/bachya/aioflo)\n[![Maintainability](https://api.codeclimate.com/v1/badges/1b6949e0c97708925315/maintainability)](https://codeclimate.com/github/bachya/aioflo/maintainability)\n[![Say Thanks](https://img.shields.io/badge/SayThanks-!-1EAEDB.svg)](https://saythanks.io/to/bachya)\n\n`aioflo` is a Python 3, `asyncio`-friendly library for interacting with\n[Flo by Moen Smart Water Detectors](https://www.moen.com/flo).\n\n# Python Versions\n\n`aioflo` is currently supported on:\n\n* Python 3.6\n* Python 3.7\n* Python 3.8\n* Python 3.9\n* Python 3.10\n\n# Installation\n\n```python\npip install aioflo\n```\n\n# Usage\n\n```python\nimport asyncio\n\nfrom aiohttp import ClientSession\n\nfrom aioflo import async_get_api\n\n\nasync def main() -> None:\n    """Run!"""\n    api = await async_get_api("<EMAIL>", "<PASSWORD>")\n\n    # Get user account information:\n    user_info = await api.user.get_info()\n    a_location_id = user_info["locations"][0]["id"]\n\n    # Get location (i.e., device) information:\n    location_info = await api.location.get_info(a_location_id)\n\n    # Get device information\n    first_device_id = location_info["devices"][0]["id"]\n    device_info = await api.device.get_info(first_device_id)\n\n    # Run a health test\n    health_test_response = await api.device.run_health_test(first_device_id)\n\n    # Close the shutoff valve\n    close_valve_response = await api.device.close_valve(first_device_id)\n\n    # Open the shutoff valve\n    open_valve_response = await api.device.open_valve(first_device_id)\n\n    # Get consumption info between a start and end datetime:\n    consumption_info = await api.water.get_consumption_info(\n        a_location_id,\n        datetime(2020, 1, 16, 0, 0),\n        datetime(2020, 1, 16, 23, 59, 59, 999000),\n    )\n\n    # Get various other metrics related to water usage:\n    metrics = await api.water.get_metrics(\n        "<DEVICE_MAC_ADDRESS>",\n        datetime(2020, 1, 16, 0, 0),\n        datetime(2020, 1, 16, 23, 59, 59, 999000),\n    )\n\n    # Set the device in "Away" mode:\n    await set_mode_away(a_location_id)\n\n    # Set the device in "Home" mode:\n    await set_mode_home(a_location_id)\n\n    # Set the device in "Sleep" mode for 120 minutes, then return to "Away" mode:\n    await set_mode_sleep(a_location_id, 120, "away")\n\n\nasyncio.run(main())\n```\n\nBy default, the library creates a new connection to Flo with each coroutine. If you are\ncalling a large number of coroutines (or merely want to squeeze out every second of\nruntime savings possible), an\n[`aiohttp`](https://github.com/aio-libs/aiohttp) `ClientSession` can be used for connection\npooling:\n\n```python\nimport asyncio\n\nfrom aiohttp import ClientSession\n\nfrom aioflo import async_get_api\n\n\nasync def main() -> None:\n    """Create the aiohttp session and run the example."""\n    async with ClientSession() as websession:\n        api = await async_get_api("<EMAIL>", "<PASSWORD>", session=session)\n\n        # Get user account information:\n        user_info = await api.user.get_info()\n        a_location_id = user_info["locations"][0]["id"]\n\n        # Get location (i.e., device) information:\n        location_info = await api.location.get_info(a_location_id)\n\n        # Get device information\n        first_device_id = location_info["devices"][0]["id"]\n        device_info = await api.device.get_info(first_device_id)\n\n        # Run a health test\n        health_test_response = await api.device.run_health_test(first_device_id)\n\n        # Close the shutoff valve\n        close_valve_response = await api.device.close_valve(first_device_id)\n\n        # Open the shutoff valve\n        open_valve_response = await api.device.open_valve(first_device_id)\n\n        # Get consumption info between a start and end datetime:\n        consumption_info = await api.water.get_consumption_info(\n            a_location_id,\n            datetime(2020, 1, 16, 0, 0),\n            datetime(2020, 1, 16, 23, 59, 59, 999000),\n        )\n\n        # Get various other metrics related to water usage:\n        metrics = await api.water.get_metrics(\n            "<DEVICE_MAC_ADDRESS>",\n            datetime(2020, 1, 16, 0, 0),\n            datetime(2020, 1, 16, 23, 59, 59, 999000),\n        )\n\n        # Set the device in "Away" mode:\n        await set_mode_away(a_location_id)\n\n        # Set the device in "Home" mode:\n        await set_mode_home(a_location_id)\n\n        # Set the device in "Sleep" mode for 120 minutes, then return to "Away" mode:\n        await set_mode_sleep(a_location_id, 120, "away")\n\n\nasyncio.run(main())\n```\n\n# Contributing\n\n1. [Check for open features/bugs](https://github.com/bachya/aioflo/issues)\n  or [initiate a discussion on one](https://github.com/bachya/aioflo/issues/new).\n2. [Fork the repository](https://github.com/bachya/aioflo/fork).\n3. (_optional, but highly recommended_) Create a virtual environment: `python3 -m venv .venv`\n4. (_optional, but highly recommended_) Enter the virtual environment: `source ./.venv/bin/activate`\n5. Install the dev environment: `script/setup`\n6. Code your new feature or bug fix.\n7. Write tests that cover your new functionality.\n8. Run tests and ensure 100% code coverage: `script/test`\n9. Update `README.md` with any new documentation.\n10. Add yourself to `AUTHORS.md`.\n11. Submit a pull request!\n',
    'author': 'Aaron Bach',
    'author_email': 'bachya1208@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bachya/aioflo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.0,<4.0.0',
}


setup(**setup_kwargs)



"""Define /alarms endpoints."""
from typing import Awaitable, Callable

from .const import API_V2_BASE


class Alarm:  # pylint: disable=too-few-public-methods
    """Define an object to handle the endpoints."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._request: Callable[..., Awaitable] = request

    async def get_all(self):
        """Get all alarms."""
        return await self._request("get", f"{API_V2_BASE}/alarms")
